@@ -1,29 +1,20 @@
-from ast import literal_eval
 from datetime import date, datetime
-from functools import lru_cache, partial
+from functools import partial
 from getpass import getpass
 from logging import Logger
 from sqlite3 import Cursor
-from typing import (
-    Any,
-    Collection,
-    Mapping,
-    Sequence,
-    Union,
-    get_args as typing_get_args,
-)
+from typing import Any, Collection
+from typing import Mapping, Sequence, Union
+from typing import get_args as typing_get_args
 
 import psycopg2
+from psycopg2._psycopg import connection
 from pyproj import CRS
-from shapely import wkb, wkt
-from shapely.geometry import shape as shapely_shape
 from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry, GEOMETRY_TYPES
 from sshtunnel import SSHTunnelForwarder
 
-from ..table import (
-    DatabaseTable,
-    random_open_tcp_port,
-)
+from tablecrow.tables.base import DatabaseTable, parse_record_values, random_open_tcp_port
+
 from ..utilities import parse_hostname, split_hostname_port
 
 SSH_DEFAULT_PORT = 22
@@ -60,6 +51,8 @@ class PostGresTable(DatabaseTable):
         logger: Logger = None,
         **kwargs,
     ):
+        self.tunnel_credentials = {}
+
         if 'ssh_hostname' in kwargs and kwargs['ssh_hostname'] is not None:
             credentials = parse_hostname(kwargs['ssh_hostname'])
             ssh_hostname = credentials['hostname']
@@ -70,23 +63,10 @@ class PostGresTable(DatabaseTable):
             ssh_username = kwargs['ssh_username'] if 'ssh_username' in kwargs else None
             ssh_password = kwargs['ssh_password'] if 'ssh_password' in kwargs else None
 
-            port = split_hostname_port(hostname)[-1]
-            if port is None:
-                port = self.DEFAULT_PORT
-
-            self.tunnel = SSHTunnelForwarder(
-                (ssh_hostname, ssh_port),
-                ssh_username=ssh_username,
-                ssh_password=ssh_password,
-                remote_bind_address=('localhost', port),
-                local_bind_address=('localhost', random_open_tcp_port()),
-            )
-            try:
-                self.tunnel.start()
-            except Exception as error:
-                raise ConnectionError(error)
-        else:
-            self.tunnel = None
+            self.tunnel_credentials['ssh_hostname'] = ssh_hostname
+            self.tunnel_credentials['ssh_port'] = ssh_port
+            self.tunnel_credentials['ssh_username'] = ssh_username
+            self.tunnel_credentials['ssh_password'] = ssh_password
 
         password = None
         if username is not None and ':' in username:
@@ -119,15 +99,16 @@ class PostGresTable(DatabaseTable):
             )
 
         if self.fields is None:
-            with self.connection:
-                with self.connection.cursor() as cursor:
+            with self.connection as connection:
+                with connection.cursor() as cursor:
                     self._DatabaseTable__fields = database_table_fields(cursor, self.name)
+            connection.close()
 
             if self.primary_key is None:
                 self._DatabaseTable__primary_key = list(self.fields)[0]
 
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 if database_has_table(cursor, self.name):
                     if database_table_is_inherited(cursor, self.name):
                         raise RuntimeError(
@@ -213,6 +194,7 @@ class PostGresTable(DatabaseTable):
                         cursor.execute(
                             f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.name} TO {user};'
                         )
+        connection.close()
         if 'password' in kwargs:
             kwargs['password'] = '*****'
         self.kwargs = kwargs
@@ -222,8 +204,28 @@ class PostGresTable(DatabaseTable):
         return self.resource
 
     @property
-    @lru_cache(maxsize=None)
-    def connection(self) -> psycopg2.connect:
+    def tunnel(self) -> SSHTunnelForwarder:
+        if 'ssh_hostname' in self.tunnel_credentials:
+            port = split_hostname_port(self.resource)[-1]
+            if port is None:
+                port = self.DEFAULT_PORT
+            tunnel = SSHTunnelForwarder(
+                (self.tunnel_credentials['ssh_hostname'], self.tunnel_credentials['ssh_port']),
+                ssh_username=self.tunnel_credentials['ssh_username'],
+                ssh_password=self.tunnel_credentials['ssh_password'],
+                remote_bind_address=('localhost', port),
+                local_bind_address=('localhost', random_open_tcp_port()),
+            )
+            try:
+                tunnel.start()
+            except Exception as error:
+                raise ConnectionError(error)
+        else:
+            tunnel = None
+        return tunnel
+
+    @property
+    def connection(self) -> connection:
         connector = partial(
             psycopg2.connect,
             database=self.database,
@@ -231,10 +233,9 @@ class PostGresTable(DatabaseTable):
             password=self._DatabaseTable__password,
         )
 
-        if self.tunnel is not None:
-            connection = connector(
-                host=self.tunnel.local_bind_host, port=self.tunnel.local_bind_port
-            )
+        tunnel = self.tunnel
+        if tunnel is not None:
+            connection = connector(host=tunnel.local_bind_host, port=tunnel.local_bind_port)
         else:
             connection = connector(host=self.hostname, port=self.port)
 
@@ -242,9 +243,11 @@ class PostGresTable(DatabaseTable):
 
     @property
     def exists(self) -> bool:
-        with self.connection:
-            with self.connection.cursor() as cursor:
-                return database_has_table(cursor, self.name)
+        with self.connection as connection:
+            with connection.cursor() as cursor:
+                exists = database_has_table(cursor, self.name)
+        connection.close()
+        return exists
 
     @property
     def schema(self) -> str:
@@ -282,8 +285,9 @@ class PostGresTable(DatabaseTable):
                 f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
             )
 
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        fields = None
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 if database_has_table(cursor, self.name):
                     fields = database_table_fields(cursor, self.name)
 
@@ -317,19 +321,22 @@ class PostGresTable(DatabaseTable):
                         fields[field] = field_type
                 else:
                     fields = None
+        connection.close()
 
-                return fields
+        return fields
 
     @property
     def connected(self) -> bool:
-        with self.connection:
+        with self.connection as connection:
             try:
-                with self.connection.cursor() as cursor:
+                with connection.cursor() as cursor:
                     cursor.execute('SELECT 1;')
                     cursor.fetchone()
-                return True
+                connected = True
             except:
-                return False
+                connected = False
+        connection.close()
+        return connected
 
     def records_where(
         self, where: Union[Mapping[str, Any], str, Sequence[str]]
@@ -341,8 +348,8 @@ class PostGresTable(DatabaseTable):
 
         where_clause, where_values = self.__where_clause(where)
 
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 if where_clause is None:
                     cursor.execute(f'SELECT {", ".join(self.fields.keys())} FROM {self.name};')
                 else:
@@ -355,6 +362,7 @@ class PostGresTable(DatabaseTable):
                     except psycopg2.errors.SyntaxError as error:
                         raise SyntaxError(f'invalid SQL syntax - {error}')
                 matching_records = cursor.fetchall()
+        connection.close()
 
         matching_records = [
             parse_record_values(dict(zip(self.fields.keys(), record)), self.fields)
@@ -386,12 +394,13 @@ class PostGresTable(DatabaseTable):
             where_clause.append(f'ST_Intersects({field}, {geometry_string})')
         where_clause = ' OR '.join(where_clause)
 
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 cursor.execute(
                     f'SELECT * FROM {self.name} WHERE {where_clause};', where_values
                 )
                 records = cursor.fetchall()
+        connection.close()
 
         return [
             parse_record_values(dict(zip(self.fields.keys(), record)), self.fields)
@@ -412,8 +421,8 @@ class PostGresTable(DatabaseTable):
                 f'no connection to {self.username}@{self.hostname}:{self.port}/{self.database}/{self.name}'
             )
 
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 for record in records:
                     if len(self.primary_key) == 1:
                         primary_key_string = self.primary_key[0]
@@ -451,6 +460,12 @@ class PostGresTable(DatabaseTable):
                         if field not in geometry_fields
                     ]
 
+                    for index, value in enumerate(values):
+                        if isinstance(value, Collection) and not isinstance(
+                            value, (str, list)
+                        ):
+                            values[index] = list(value)
+
                     if primary_key_value in self:
                         record_without_primary_key = {
                             column: value
@@ -495,6 +510,7 @@ class PostGresTable(DatabaseTable):
                                 f'WHERE {primary_key_string} = %s;',
                                 [geometry.wkt, self.crs.to_epsg(), primary_key_value],
                             )
+        connection.close()
 
     def delete_where(self, where: Union[Mapping[str, Any], str, Sequence[str]]):
         if not self.connected:
@@ -504,8 +520,8 @@ class PostGresTable(DatabaseTable):
 
         where_clause, where_values = self.__where_clause(where)
 
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 if where_clause is None:
                     cursor.execute(f'TRUNCATE {self.name};')
                 else:
@@ -517,17 +533,21 @@ class PostGresTable(DatabaseTable):
                         raise KeyError(error)
                     except psycopg2.errors.SyntaxError as error:
                         raise SyntaxError(f'invalid SQL syntax - {error}')
+        connection.close()
 
     def __len__(self) -> int:
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 cursor.execute(f'SELECT COUNT(*) FROM {self.name};')
-                return cursor.fetchone()[0]
+                length = cursor.fetchone()[0]
+        connection.close()
+        return length
 
     def delete_table(self):
-        with self.connection:
-            with self.connection.cursor() as cursor:
+        with self.connection as connection:
+            with connection.cursor() as cursor:
                 cursor.execute(f'DROP TABLE {self.name};')
+        connection.close()
 
     def __repr__(self) -> str:
         return (
@@ -538,10 +558,6 @@ class PostGresTable(DatabaseTable):
             f'{", ".join(key + "=" + repr(value) for key, value in self.kwargs.items())})'
         )
 
-    def __del__(self):
-        if self.tunnel is not None:
-            self.tunnel.stop()
-
     def __where_clause(self, where: {str: Union[Any, list]}) -> (str, [Any]):
         if (
             where is not None
@@ -573,9 +589,10 @@ class PostGresTable(DatabaseTable):
                                 statement = f'%s = ANY({field})'
                             else:
                                 if fields is None:
-                                    with self.connection:
-                                        with self.connection.cursor() as cursor:
+                                    with self.connection as connection:
+                                        with connection.cursor() as cursor:
                                             fields = database_table_fields(cursor, self.name)
+                                    connection.close()
                                 field_type = fields[field]
                                 dimensions = field_type.count('_')
                                 field_type = field_type.strip('_')
@@ -605,55 +622,6 @@ class PostGresTable(DatabaseTable):
         return where_clause, where_values
 
 
-def parse_record_values(record: {str: Any}, field_types: {str: type}) -> {str: Any}:
-    """
-    Parse the values in the given record into their respective field types.
-
-    :param record: dictionary mapping fields to values
-    :param field_types: dictionary mapping fields to types
-    :return: record with values parsed into their respective types
-    """
-
-    for field, value in record.items():
-        if field in field_types:
-            field_type = field_types[field]
-            value_type = type(value)
-
-            if value_type is not field_type and value is not None:
-                if field_type is bool:
-                    value = (
-                        bool(value)
-                        if value_type is not str
-                        else literal_eval(value.capitalize())
-                    )
-                elif field_type is int:
-                    value = int(value)
-                elif field_type is float:
-                    value = float(value)
-                elif field_type is str:
-                    value = str(value)
-                elif value_type in (str, bytes):
-                    if field_type is list:
-                        value = literal_eval(value)
-                    elif field_type is datetime:
-                        value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
-                    elif field_type is date:
-                        value = datetime.strptime(value, '%Y-%m-%d').date()
-                    elif field_type.__name__ in GEOMETRY_TYPES:
-                        try:
-                            value = wkb.loads(value, hex=True)
-                        except:
-                            try:
-                                value = wkt.loads(value)
-                            except:
-                                try:
-                                    value = wkb.loads(value)
-                                except TypeError:
-                                    value = shapely_shape(literal_eval(value))
-                record[field] = value
-    return record
-
-
 def database_tables(cursor: Cursor, user_defined: bool = True) -> [str]:
     """
     List tables within the given database.
